//this file is part of eMule
//Copyright (C)2002-2006 Merkur ( strEmail.Format("%s@%s", "devteam", "emule-project.net") / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "eMule.h"
#include "HTRichEditCtrl.h"
#include "TitleMenu.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

IMPLEMENT_DYNAMIC(CHTRichEditCtrl, CRichEditCtrl)

BEGIN_MESSAGE_MAP(CHTRichEditCtrl, CRichEditCtrl)
	ON_WM_CONTEXTMENU()
	ON_WM_LBUTTONDOWN()
	ON_WM_SETCURSOR()
	ON_WM_KEYDOWN()
	ON_NOTIFY_REFLECT_EX(EN_LINK, OnEnLink)
	ON_WM_CREATE()
END_MESSAGE_MAP()

CHTRichEditCtrl::CHTRichEditCtrl() : m_strTitle(_T("")), m_strURL(_T(""))
{
	m_dwFlags = (HTC_ISAUTOSCROLL | HTC_ISDEFAULTLINKS);
	m_crDefaultForeground = GetSysColor(COLOR_WINDOWTEXT);
	m_crDefaultBackground = GetSysColor(COLOR_WINDOW);
	switch (get_cpu_type())
	{
	case 1:
		m_lMaxBufSize = (32 * 1024);
		break;
	case 2:
		m_lMaxBufSize = (64 * 1024);
		break;
	default:
		if (g_App.m_pPrefs->GetWindowsVersion() == _WINVER_95_)
			m_lMaxBufSize = (64 * 1024);
		else
			m_lMaxBufSize = (128 * 1024);
		break;
	}
	m_hArrowCursor = ::LoadCursor(NULL, IDC_ARROW);
	m_hHandCursor = ::LoadCursor(NULL, MAKEINTRESOURCE(32649));
}

CHTRichEditCtrl::~CHTRichEditCtrl()
{
	//	Shared cursors mustn't be destroyed
}

void CHTRichEditCtrl::SetTitle(LPCTSTR pszTitle)
{
	m_strTitle = g_App.StripInvalidFilenameChars(pszTitle);
}

int CHTRichEditCtrl::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CRichEditCtrl::OnCreate(lpCreateStruct) == -1)
		return -1;

	VERIFY( SendMessage(EM_SETUNDOLIMIT, 0, 0) == 0 );

	LimitText((g_App.m_pPrefs->GetWindowsVersion() == _WINVER_95_) ? ((64 * 1024) - 1) : (128 * 1024));
	SetTargetDevice(NULL, (bool)(m_dwFlags & HTC_ISWORDWRAP) == 0);
	SetEventMask(GetEventMask() | ENM_LINK);

	return 0;
}

void CHTRichEditCtrl::AppendText(LPCTSTR pszMsg, COLORREF crTextColor /*= CLR_DEFAULT*/, COLORREF crBackColor /*= CLR_DEFAULT*/, DWORD dwFlags /*= 0*/)
{
	long		lMsgLen = _tcslen(pszMsg);

	if (lMsgLen == 0)
		return;

	long		lStartChar, lEndChar, lTextEnd = GetTextLengthEx(GTL_PRECISE);

	GetSel(lStartChar, lEndChar);

	bool			bIsVisible = IsWindowVisible(), bIsEndSel = (lStartChar == lEndChar && lStartChar == lTextEnd);
	CHARFORMAT2		cf2;
	POINT			ptOldPos;

	if (bIsVisible)
		SetRedraw(false);

	if (lTextEnd >= m_lMaxBufSize)
	{
		CString		strLog;

		GetWindowText(strLog);

		long		lLen = strLog.Find(_T('\n'), m_lMaxBufSize >> 3);

		if (lLen == -1)
			lLen = m_lMaxBufSize >> 3;

		strLog.Truncate(lLen);
		strLog.Remove(_T('\n'));
		lLen = strLog.GetLength();
		SetSel(0, lLen);
		ReplaceSel(_T(""));
	}
	SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptOldPos);
	SetSel(-1, -1);
	memzero(&cf2, sizeof cf2);
	cf2.cbSize = sizeof cf2;
	GetSelectionCharFormat(cf2);
	cf2.dwEffects &= ~(CFE_AUTOCOLOR | CFE_AUTOBACKCOLOR | CFE_BOLD | CFE_ITALIC | CFE_LINK | CFE_STRIKEOUT | CFE_UNDERLINE);
	cf2.crTextColor = (crTextColor == CLR_DEFAULT) ? m_crDefaultForeground : crTextColor;
	cf2.crBackColor = (crBackColor == CLR_DEFAULT) ? m_crDefaultBackground : crBackColor;
	if ((dwFlags & HTC_BOLD) != 0)
		cf2.dwEffects |= CFE_BOLD;
	if ((dwFlags & HTC_ITALIC) != 0)
		cf2.dwEffects |= CFE_ITALIC;
	if ((dwFlags & HTC_LINK) != 0)
		cf2.dwEffects |= CFE_LINK;
	if ((dwFlags & HTC_STRIKEOUT) != 0)
		cf2.dwEffects |= CFE_STRIKEOUT;
	if ((dwFlags & HTC_UNDERLINE) != 0)
		cf2.dwEffects |= CFE_UNDERLINE;
	SetSelectionCharFormat(cf2);
	ReplaceSel(pszMsg);

	if ((dwFlags & HTC_LINK) == 0 && (m_dwFlags & HTC_ISDEFAULTLINKS) != 0)
	{
		CString		strLineOrginal = pszMsg, strLine = strLineOrginal;
		int			iIndex = 0, iScheme, iLen, iLinkEnd;

		lTextEnd = GetTextLengthEx(GTL_PRECISE) - strLineOrginal.GetLength();
		cf2.dwEffects |= CFE_LINK;

		//	Convert string to lower case only once
		strLine.MakeLower();
		while (strLineOrginal.GetLength() > iIndex)
		{
			for (iScheme = 0; iScheme < ARRSIZE(_apszSchemes); iScheme++)
			{
				iLen = _apszSchemes[iScheme].uiLen;
				iLinkEnd = iIndex + iLen;

				if ((strLine.GetLength() > iLinkEnd) &&
					(strLine.Mid(iIndex, iLen).Compare(_apszSchemes[iScheme].pszScheme) == 0))
				{
					while (strLine.GetLength() > iLinkEnd && strLine[iLinkEnd] != _T(' ')
						&& strLine[iLinkEnd] != _T('\n') && strLine[iLinkEnd] != _T('\\') &&
						strLine[iLinkEnd] != _T('\r') && strLine[iLinkEnd] != _T('\t'))
					{
						iLinkEnd++;
					}

					while ((iLinkEnd > iIndex + iLen) && 
						!((strLine[iLinkEnd - 1] >= _T('/') && strLine[iLinkEnd - 1] <= _T('9'))
						||(strLine[iLinkEnd - 1] >= _T('a') && strLine[iLinkEnd - 1] <= _T('z'))
						|| (strLine[iLinkEnd - 1] == _T('|'))))
					{
						iLinkEnd--;
					}

					if (iLinkEnd > iIndex + iLen)
					{
						SetSel(lTextEnd + iIndex, lTextEnd + iLinkEnd);
						SetSelectionCharFormat(cf2);
						iIndex = iLinkEnd;
					}
					else
						iIndex += iLen;

					break;
				}
			}
			if (iScheme >= ARRSIZE(_apszSchemes))
				iIndex++;
		}
	}

	if (bIsEndSel)
		SetSel(-1, -1);
	else
		SetSel(lStartChar, lEndChar);

	if ((m_dwFlags & HTC_ISAUTOSCROLL) != 0 && (_tcschr(pszMsg, _T('\n')) != NULL))
	{
		ScrollToLastLine();

		POINT			ptNewPos;

		SendMessage(EM_GETSCROLLPOS, 0, (LPARAM)&ptNewPos);
		ptOldPos.y = ptNewPos.y;
	}

	if ((m_dwFlags & HTC_ISWORDWRAP) != 0)
		ptOldPos.x = 0;

	SendMessage(EM_SETSCROLLPOS, 0, (LPARAM)&ptOldPos);

	if (bIsVisible)
	{
		SetRedraw();
		Invalidate();
	}
}

void CHTRichEditCtrl::ScrollToLastLine()
{
	long		lFirstVisible = GetFirstVisibleLine();

	if (lFirstVisible > 0)
		LineScroll(-lFirstVisible, 0);

	// WM_VSCROLL does not work correctly under Win98 (or older versions of comctl.dll)
	SendMessage(WM_VSCROLL, SB_BOTTOM);
	SendMessage(WM_VSCROLL, SB_LINEUP);
	if (g_App.m_pPrefs->GetWindowsVersion() == _WINVER_95_)
	{
		// older versions of comctl.dll seem to need this to properly update the display
		SendMessage(WM_VSCROLL, MAKELONG(SB_THUMBPOSITION, GetScrollPos(SB_VERT)));
		SendMessage(WM_VSCROLL, SB_ENDSCROLL);
	}
}

void CHTRichEditCtrl::Reset()
{
	SetRedraw(false);
	SetWindowText(_T(""));
	SetRedraw();
	Invalidate();
}

void CHTRichEditCtrl::OnContextMenu(CWnd *pWnd, CPoint point)
{
	long			lSelStart, lSelEnd;
	CTitleMenu		menuLog;

	GetSel(lSelStart, lSelEnd);
	m_dwFlags |= HTC_ISARROWCURSOR;
	menuLog.CreatePopupMenu();
	if ((m_dwFlags & HTC_ISLIMITED) == 0)
		menuLog.AddMenuTitle(GetResString(IDS_LOGENTRY));
	menuLog.AppendMenu(MF_STRING | ((lSelEnd > lSelStart) ? MF_ENABLED : MF_GRAYED), MP_COPYSELECTED, GetResString(IDS_COPY));

	UINT			dwFlags = MF_STRING | ((GetWindowTextLength() > 0) ? MF_ENABLED : MF_GRAYED);
	CString			strBuffer = GetResString(IDS_SAVETOFILE);

	if ((m_dwFlags & HTC_ISLIMITED) == 0)
	{
		menuLog.AppendMenu(dwFlags, MP_SAVELOG, strBuffer + _T(" (.log)"));
		menuLog.AppendMenu(dwFlags, MP_SAVERTF, strBuffer + _T(" (.rtf)"));
	}
	menuLog.AppendMenu(MF_SEPARATOR);
	menuLog.AppendMenu(dwFlags, MP_SELECTALL, GetResString(IDS_SELECTALL));

	if ((m_dwFlags & HTC_ISLIMITED) == 0)
	{
		menuLog.AppendMenu(dwFlags, MP_REMOVEALL, GetResString(IDS_PW_RESET));
		menuLog.AppendMenu(MF_SEPARATOR);
		menuLog.AppendMenu(MF_STRING | ((m_dwFlags & HTC_ISAUTOSCROLL) != 0 ? MF_CHECKED : MF_UNCHECKED), MP_AUTOSCROLL, GetResString(IDS_AUTOSCROLL));
		menuLog.AppendMenu(MF_STRING | ((m_dwFlags & HTC_ISWORDWRAP) != 0 ? MF_CHECKED : MF_UNCHECKED), MP_WORDWRAP, GetResString(IDS_WORDWRAP));
	}

	menuLog.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);
	m_dwFlags &= ~HTC_ISARROWCURSOR;
}

BOOL CHTRichEditCtrl::OnCommand(WPARAM wParam, LPARAM lParam)
{
	switch (wParam)
	{
	case MP_COPYSELECTED:
		Copy();
		break;
	case MP_SAVELOG:
		SaveLogToDisk();
		break;
	case MP_SAVERTF:
		SaveRtfToDisk();
		break;
	case MP_SELECTALL:
		SetSel(0, -1);
		break;
	case MP_REMOVEALL:
		Reset();
		break;
	case MP_AUTOSCROLL:
		m_dwFlags ^= HTC_ISAUTOSCROLL;
		break;
	case MP_WORDWRAP:
		m_dwFlags ^= HTC_ISWORDWRAP;
		SetTargetDevice(NULL, (bool)(m_dwFlags & HTC_ISWORDWRAP) == 0);
		break;
	}
	return true;
}

void CHTRichEditCtrl::OnLButtonDown(UINT nFlags, CPoint point)
{
	if ((m_dwFlags & HTC_ISDEFAULTLINKS) == 0 && (m_dwFlags & HTC_ISHANDCURSOR) != 0 && !m_strURL.IsEmpty())
		ShellExecute(NULL, NULL, m_strURL, NULL, NULL, SW_SHOWDEFAULT);
	else
		CRichEditCtrl::OnLButtonDown(nFlags, point);
}

BOOL CHTRichEditCtrl::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
	if ((m_dwFlags & HTC_ISARROWCURSOR) != 0)
	{
		::SetCursor(m_hArrowCursor);
		return true;
	}
	else if ((m_dwFlags & HTC_ISHANDCURSOR) != 0)
	{
		::SetCursor(m_hHandCursor);
		return true;
	}

	return CRichEditCtrl::OnSetCursor(pWnd, nHitTest, message);
}

void CHTRichEditCtrl::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if (GetKeyState(VK_CONTROL) & 0x8000)
	{
		if (nChar == _T('A'))
			//////////////////////////////////////////////////////////////////
			// Ctrl+A: Select all items
			SetSel(0, -1);
		else if (nChar == _T('C'))
			//////////////////////////////////////////////////////////////////
			// Ctrl+C: Copy listview items to clipboard
			Copy();
	}
	else if (nChar == VK_ESCAPE)
		return;	// don't minimize CHTRichEditCtrl

	CRichEditCtrl::OnKeyDown(nChar, nRepCnt, nFlags);
}

BOOL CHTRichEditCtrl::OnEnLink(NMHDR *pNMHDR, LRESULT *pResult)
{
	ENLINK		*pEnLink = reinterpret_cast<ENLINK *>(pNMHDR);

	if (pEnLink != NULL)
	{
		CString		strURL;

		GetTextRange(pEnLink->chrg.cpMin, pEnLink->chrg.cpMax, strURL);

		for (int iScheme = 0; iScheme < ARRSIZE(_apszSchemes); iScheme++)
		{
			if (strURL.Left(_apszSchemes[iScheme].uiLen).CompareNoCase(_apszSchemes[iScheme].pszScheme) == 0)
			{
				m_strURL = strURL;
				
				if (pEnLink->msg == WM_LBUTTONDOWN)
				{
					ShellExecute(NULL, NULL, m_strURL, NULL, NULL, SW_SHOWDEFAULT);
					*pResult = 1;
					return true;
				}
			}
		}
	}

	*pResult = 0;
	return false;
}

void CHTRichEditCtrl::SaveLogToDisk()
{
	CString			strLog, strTemp;

	if (DialogBrowseFile(strTemp, _T("Log (*.log)|*.log||"), m_strTitle + _T(".log"), 0, false))
	{
		HANDLE		hFile = CreateFile(strTemp, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			GetWindowText(strLog);

			DWORD		dwWritten;

			WriteFile(hFile, strLog, strLog.GetLength(), &dwWritten, NULL);
			CloseHandle(hFile);
		}
	}
}

void CHTRichEditCtrl::SaveRtfToDisk()
{
	CString			strTemp;

	if (DialogBrowseFile(strTemp, _T("Rich Text Format (RTF)|*.rtf||"), m_strTitle + _T(".rtf"), 0, false))
	{
		HANDLE		hFile = CreateFile(strTemp, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			EDITSTREAM		es;
			CString			strRTF;
			DWORD			dwWritten;

			es.dwCookie = (DWORD)&strRTF;
			es.pfnCallback = MEditStreamOutCallback;
			StreamOut(SF_RTF, es);
			WriteFile(hFile, strRTF, strRTF.GetLength(), &dwWritten, NULL);
			CloseHandle(hFile);
		}
	}
}

CString CHTRichEditCtrl::GetHtml()
{
	CString			strTemp, strText;
	EDITSTREAM		es;

	es.dwCookie = (DWORD)&strTemp;
	es.pfnCallback = MEditStreamOutCallback;
	StreamOut(SF_RTF, es);
	strTemp.Replace(_T("&"), _T("&amp;"));
	strTemp.Replace(_T("<"), _T("&lt;"));
	strTemp.Replace(_T(">"), _T("&gt;"));
	strTemp.Replace(_T("\""), _T("&quot;"));
	strTemp.Replace(_T("\\\'"), _T(" \\\'"));

	int				iTableStart = strTemp.Find(_T("\n{\\colortbl ;"));
	int				iTableEnd = strTemp.Find(_T(";}"), iTableStart + 1);

	if ((iTableStart != -1) && (iTableEnd != -1))
	{
		CStringArray	astrColorsRtf, astrColorsHtml;
		int				iRtfColor = 0, iColorEnd;
		long			lHtmlColor = 0;
		CString			strColor;

		while ((iTableStart < iTableEnd) && (iTableStart != -1))
		{
			iTableStart = strTemp.Find(_T("\\red"), iTableStart + 1);
			strColor.Format(_T("\\cf%i"), iRtfColor);
			astrColorsRtf.Add(strColor);
			strColor.Format(_T(" </font><font color=\"#%06x\">"), lHtmlColor);
			astrColorsHtml.Add(strColor);
			iRtfColor ++;
			if (iTableStart != -1)
			{
				iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
				if (iColorEnd != -1)
				{
					lHtmlColor = (_tstoi(strTemp.Mid(iTableStart + 4, iColorEnd - (iTableStart + 4))) << 16);
					iTableStart = iColorEnd + 6;
					iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
					if (iColorEnd != -1)
					{
						lHtmlColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 8);
						iTableStart = iColorEnd + 5;
						iColorEnd = strTemp.Find(_T(';'), iTableStart + 1);
						if (iColorEnd != -1)
							lHtmlColor += _tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart));
					}
				}
			}
		}
		if (astrColorsRtf.GetSize() > 0)
		{
			for (int i = 0; i < astrColorsRtf.GetSize(); i++)
			{
				strTemp.Replace(astrColorsRtf[i] + _T(" \\\'"), astrColorsHtml[i] + _T(" \\\'"));
				strTemp.Replace(astrColorsRtf[i] + _T(' '), astrColorsHtml[i]);
				strTemp.Replace(astrColorsRtf[i] + _T('\\'), astrColorsHtml[i] + _T('\\'));
			}
		}

		LPCTSTR			psz = strTemp;

		psz += strTemp.Find(_T('\\'), iTableEnd);

		LPCTSTR			pszStart = psz;
		TCHAR			strChar[2];
		long			lLen;

		while (*psz != _T('\0'))
		{
			if (*psz == _T('\\'))
			{
				if (*(psz + 1) == _T('\\'))
				{
					strText += CString(pszStart, 1 + psz - pszStart);
					psz += 2;
				}
				else if (*(psz + 1) == _T('\''))
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart - 1);

					strChar[0] = *(psz + 2);
					strChar[1] = *(psz + 3);
					strText += (TCHAR)_tcstol(strChar, NULL, 16);
					psz += 4;
				}
				else
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart);

					lLen = _tcscspn(psz, _T(" \n"));
					if (lLen == 0)
						psz += _tcslen(psz);
					else
					{
						psz += lLen;
						if (*psz == _T(' '))
							psz++;
					}
				}
				pszStart = psz;
			}
			else
				psz++;
		}
		if (pszStart != psz)
			strText += pszStart;
	}

	iTableEnd = strText.ReverseFind(_T('}'));
	if (iTableEnd != -1)
		strText.Truncate(iTableEnd);
	strText.TrimRight(_T('\n'));
	strText.Replace(_T(" </font><font color"), _T("</font><font color"));

	return strText;
}

CString CHTRichEditCtrl::GetToolTip()
{
	CString			strTemp, strText;
	EDITSTREAM		es;

	es.dwCookie = (DWORD)&strTemp;
	es.pfnCallback = MEditStreamOutCallback;
	StreamOut(SF_RTF, es);
	strTemp.Replace(_T("<"), _T("<<"));
	strTemp.Replace(_T("\\\'"), _T(" \\\'"));

	int				iTableStart = strTemp.Find(_T("\n{\\colortbl ;"));
	int				iTableEnd = strTemp.Find(_T(";}"), iTableStart + 1);

	if ((iTableStart != -1) && (iTableEnd != -1))
	{
		CStringArray	astrColorsRtf, astrColorsTT;
		int				iRtfColor = 0, iColorEnd;
		long			lTTColor = 0;
		CString			strColor;

		while ((iTableStart < iTableEnd) && (iTableStart != -1))
		{
			iTableStart = strTemp.Find(_T("\\red"), iTableStart + 1);
			strColor.Format(_T("\\cf%i"), iRtfColor);
			astrColorsRtf.Add(strColor);
			(iRtfColor == 0 || lTTColor == GetSysColor(COLOR_WINDOWTEXT)) ? strColor = _T(" <ct>") : strColor.Format(_T(" <ct=0x%06x>"), lTTColor);
			astrColorsTT.Add(strColor);
			iRtfColor++;
			if (iTableStart != -1)
			{
				iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
				if (iColorEnd != -1)
				{
					lTTColor = _tstoi(strTemp.Mid(iTableStart + 4, iColorEnd - (iTableStart + 4)));
					iTableStart = iColorEnd + 6;
					iColorEnd = strTemp.Find(_T('\\'), iTableStart + 1);
					if (iColorEnd != -1)
					{
						lTTColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 8);
						iTableStart = iColorEnd + 5;
						iColorEnd = strTemp.Find(_T(';'), iTableStart + 1);
						if (iColorEnd != -1)
							lTTColor += (_tstoi(strTemp.Mid(iTableStart, iColorEnd - iTableStart)) << 16);
					}
				}
			}
		}
		if (astrColorsRtf.GetSize() > 0)
		{
			for (int i = 0; i < astrColorsRtf.GetSize(); i++)
			{
				strTemp.Replace(astrColorsRtf[i] + _T(" \\\'"), astrColorsTT[i] + _T(" \\\'"));
				strTemp.Replace(astrColorsRtf[i] + _T(' '), astrColorsTT[i]);
				strTemp.Replace(astrColorsRtf[i] + _T('\\'), astrColorsTT[i] + _T('\\'));
			}
		}

		int			iLines = 0, iStrEnd = strTemp.ReverseFind(_T('}'));

		if (iStrEnd != -1)
			strTemp.Truncate(iStrEnd);
		strTemp.TrimRight(_T('\n'));

		LPCTSTR		pszStart = strTemp;
		LPCTSTR		psz = pszStart;

		pszStart += strTemp.Find(_T('\\'), iTableEnd);
		psz += strTemp.GetLength() - 1;

		CString		strLog, strStatus = g_App.m_pMDlg->m_ctlStatusBar.GetText(SB_MESSAGETEXT, 0);
		int			iLastFound, iNextFound, iCount = 11;

		GetWindowText(strLog);
		iLastFound = strLog.Find(strStatus);
		iNextFound = iLastFound;
		while (iNextFound != -1)
		{
			iLastFound = iNextFound;
			iNextFound = strLog.Find(strStatus, iLastFound + 1);
		}
		
		if (iLastFound != -1)
		{
			iCount = 0;
			iNextFound = strLog.Find(_T("\n"), iLastFound + 1);
			while (iNextFound != -1)
			{
				iCount++;
				iNextFound = strLog.Find(_T("\n"), iNextFound + 1);
			}
			
			while (iCount > 0 && psz > pszStart)
			{
				if (*psz == _T('\n'))
					iCount--;
				psz--;
			}
			iCount = 10 - iCount;
			strTemp.Truncate(psz - strTemp);
		}

		while (iLines < iCount && psz > pszStart)
		{
			if (*psz == _T('\n'))
				iLines++;
			psz--;
		}

		if (psz > pszStart)
			psz++;

		pszStart = psz;

		TCHAR		strChar[2];
		long		lLen;

		while (*psz != _T('\0'))
		{
			if (*psz == _T('\\'))
			{
				if (*(psz + 1) == _T('\\'))
				{
					strText += CString(pszStart, 1 + psz - pszStart);
					psz += 2;
				}
				else if (*(psz + 1) == _T('\''))
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart - 1);

					strChar[0] = *(psz + 2);
					strChar[1] = *(psz + 3);
					strText += (TCHAR)_tcstol(strChar, NULL, 16);
					psz += 4;
				}
				else
				{
					if (psz - pszStart > 0)
						strText += CString(pszStart, psz - pszStart);

					lLen = _tcscspn(psz, _T(" \n"));
					if (lLen == 0)
						psz += _tcslen(psz);
					else
					{
						psz += lLen;
						if (*psz == _T(' '))
							psz++;
					}
				}
				pszStart = psz;
			}
			else
				psz++;
		}
		if (pszStart != psz)
			strText += pszStart;
	}

	strText.Trim(_T('\n'));
	strText.Replace(_T(" <ct"), _T("<ct"));
	while (strText.Right(4) == _T("<ct>"))
		strText.Truncate(strText.GetLength() - 4);
	strText.TrimRight(_T('\n'));
	strText.Replace(_T("\n"), _T("<br>"));
	if (strText[0] != _T('<'))
		strText.Insert(0, _T("<"));

	return strText;
}

void CHTRichEditCtrl::SetFont(CFont* pFont, bool bRedraw /*= true*/)
{
	SetBackgroundColor(false, m_crDefaultBackground);
	
	LOGFONT			lf = {0};
	CHARFORMAT2		cf2;

	pFont->GetLogFont(&lf);
	memzero(&cf2, sizeof cf2);
	cf2.cbSize = sizeof cf2;
	cf2.dwMask = CFM_FACE | CFM_SIZE /*| CFM_CHARSET*/;

	if (GetWindowTextLength() == 0)
		cf2.dwMask |= CFM_BACKCOLOR | CFM_BOLD | CFM_COLOR | CFM_ITALIC | CFM_LINK | CFM_STRIKEOUT | CFM_UNDERLINE;

	cf2.crTextColor = m_crDefaultForeground;
	cf2.crBackColor = m_crDefaultBackground;

	HDC				hDC = ::GetDC(NULL);

	cf2.yHeight = (-MulDiv(lf.lfHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY))) * 20;
	::ReleaseDC(NULL, hDC);

	cf2.bPitchAndFamily = lf.lfPitchAndFamily;
	_tcsncpy(cf2.szFaceName, lf.lfFaceName, ARRSIZE(cf2.szFaceName));
	cf2.szFaceName[ARRSIZE(cf2.szFaceName) - 1] = _T('\0');

	// although this should work correctly (according SDK) it may give false results (e.g. the "click here..." text
	// which is shown in the server info window may not be entirely used as a hyperlink???)
	//cf2.bCharSet = lf.lfCharSet;
	//cf2.yOffset = 0;
	VERIFY( SetDefaultCharFormat(cf2) );

	if (bRedraw)
	{
		Invalidate();
		UpdateWindow();
	}
}
