//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include "emule.h"
#include "IPFilter.h"
#include "otherfunctions.h"
#include "Preferences.h"
#include "emuleDlg.h"
#include "HttpDownloadDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


#define DFLT_FILTER_LEVEL	100 // default filter level if non specified
#define DFLT_TICKS_TO_BAN	(24u * 60u * 60u * 1000u)	//default ban time

/////////////////////////////////////////////////////////////////////////////////////////////
CIPFilter::CIPFilter() : m_strTemporaryLastHit(_T("Not available"))
{
	m_pLastHit = NULL;
}
/////////////////////////////////////////////////////////////////////////////////////////////
CIPFilter::~CIPFilter()
{
	RemoveAllIPs();
	iplist_temporary.clear();
}
/////////////////////////////////////////////////////////////////////////////////////////////
void CIPFilter::AddIPRange(uint32 dwIPFrom, uint32 dwIPTo, UINT uiLevel, const CString& rstrDesc)
{
	EMULE_TRY

	SIPFilter* newFilter = new SIPFilter;

	newFilter->dwStart	= dwIPFrom;
	newFilter->dwEnd	= dwIPTo;
	newFilter->uiLevel	= uiLevel;
	newFilter->strDesc.SetString(rstrDesc.GetString(), rstrDesc.GetLength());
#if 0
	newFilter->uiHits	= 0;
#endif

	m_IPList.Add(newFilter);

	EMULE_CATCH
}
/////////////////////////////////////////////////////////////////////////////////////////////
static int __cdecl CmpSIPFilterByStartAddr(const void* p1, const void* p2)
{
	const SIPFilter* rng1 = *(SIPFilter**)p1;
	const SIPFilter* rng2 = *(SIPFilter**)p2;

	return CompareUnsigned(rng1->dwStart, rng2->dwStart);
}
/////////////////////////////////////////////////////////////////////////////////////////////
int CIPFilter::LoadFromDefaultFile(bool bShowResponse)
{
	RemoveAllIPs();

	return AddFromFile(CString(g_App.m_pPrefs->GetConfigDir() + DFLT_IPFILTER_FILENAME), bShowResponse);
}
/////////////////////////////////////////////////////////////////////////////////////////////
int CIPFilter::AddFromFile(LPCTSTR pstrFilePath, bool bShowResponse)
{
	EMULE_TRY

	FILE	*pReadFile = _tfopen(pstrFilePath, _T("r"));

	if (pReadFile != NULL)
	{
		enum EIPFilterFileType
		{
			Unknown = 0,
			FilterDat = 1,		// ipfilter.dat/ip.prefix format
			PeerGuardian = 2	// PeerGuardian format
		} eFileType = Unknown;

		TCHAR	strNam[_MAX_FNAME];
		TCHAR	strExt[_MAX_EXT];

		_tsplitpath(pstrFilePath, NULL, NULL, strNam, strExt);
		if (_tcsicmp(strExt, _T(".p2p")) == 0 || (_tcsicmp(strNam, _T("guarding.p2p")) == 0 && _tcsicmp(strExt, _T(".txt")) == 0))
			eFileType = PeerGuardian;
		else if (_tcsicmp(strExt, _T(".prefix")) == 0)
			eFileType = FilterDat;

		int		iLine = 0;
		int		iFoundRanges = 0;
		int		iDuplicate = 0;
		int		iMerged = 0;
		CString	strBuffer;
		TCHAR	acBuffer[1024];

		setvbuf(pReadFile, NULL, _IOFBF, 64*1024);
		while (_fgetts(acBuffer, ARRSIZE(acBuffer), pReadFile) != NULL)
		{
			iLine++;
			if ((*acBuffer == _T('#')) || (*acBuffer == _T('/')))	// Ignore comments
				continue;
			strBuffer = acBuffer;
		//	Ignore too short lines
			if (strBuffer.GetLength() < 5)
				continue;

			if (eFileType == Unknown)
			{
			//	Looks like html
				if ((strBuffer.Find(_T('>')) >= 0) && (strBuffer.Find(_T('<')) >= 0))
					strBuffer.Delete(0, strBuffer.ReverseFind(_T('>')) + 1);

			//	Check for <IP> - <IP> at start of line
				UINT	u1, u2, u3, u4, u5, u6, u7, u8;

				if (_stscanf(strBuffer, _T("%u.%u.%u.%u - %u.%u.%u.%u"), &u1, &u2, &u3, &u4, &u5, &u6, &u7, &u8) == 8)
					eFileType = FilterDat;
				else
				{
				//	Check for <description> ':' <IP> '-' <IP>
					int iColon = strBuffer.Find(_T(':'));
					if (iColon >= 0)
					{
						const TCHAR	*pcIPRange = strBuffer.GetString() + iColon + 1;
						UINT		u1, u2, u3, u4, u5, u6, u7, u8;

						if (_stscanf(pcIPRange, _T("%u.%u.%u.%u - %u.%u.%u.%u"), &u1, &u2, &u3, &u4, &u5, &u6, &u7, &u8) == 8)
							eFileType = PeerGuardian;
					}
				}
			}

			bool	bValid = false;
			uint32	dwStart, dwEnd;
			UINT	uiLevel;

			if (eFileType == FilterDat)
				bValid = ParseFilterLine1(strBuffer, dwStart, dwEnd, uiLevel);
			else if (eFileType == PeerGuardian)
				bValid = ParseFilterLine2(strBuffer, dwStart, dwEnd, uiLevel);

		//	Add a filter
			if (bValid)
			{
				AddIPRange(dwStart, dwEnd, uiLevel, strBuffer);
				iFoundRanges++;
			}
		}
		fclose(pReadFile);

	//	Sort the IP filter list by IP range start addresses
		qsort(m_IPList.GetData(), m_IPList.GetCount(), sizeof(m_IPList[0]), CmpSIPFilterByStartAddr);

	//	Merge overlapping and adjacent filter ranges
		if (m_IPList.GetCount() >= 2)
		{
			SIPFilter	*pPrv = m_IPList[0];
			int	i = 1;

			while (i < m_IPList.GetCount())
			{
				SIPFilter	*pCur = m_IPList[i];

				if ( ( ((pCur->dwStart >= pPrv->dwStart) && (pCur->dwStart <= pPrv->dwEnd))		// overlapping
					|| (pCur->dwStart == pPrv->dwEnd + 1) ) && (pCur->uiLevel == pPrv->uiLevel))// adjacent
				{
					if (pCur->dwStart != pPrv->dwStart || pCur->dwEnd != pPrv->dwEnd) // don't merge identical entries
					{
					//	TODO: not yet handled, overlapping entries with different 'level'
						if (pCur->dwEnd > pPrv->dwEnd)
							pPrv->dwEnd = pCur->dwEnd;
						iMerged++;
					}
					else
					{
					//	If we have identical entries, use the lowest 'level'
						if (pCur->uiLevel < pPrv->uiLevel)
							pPrv->uiLevel = pCur->uiLevel;
						iDuplicate++;
					}
					delete pCur;
					m_IPList.RemoveAt(i);
					continue;
				}
				pPrv = pCur;
				i++;
			}
		}

		g_App.AddLogLine(bShowResponse, RGB_LOG_SUCCESS + GetResString(IDS_IPFILTERLOADED) , m_IPList.GetCount());
		if (g_App.m_pPrefs->GetVerbose())
		{
			g_App.AddDebugLogLine(_T("Loaded IP filters from \"%s\""), pstrFilePath);
			g_App.AddDebugLogLine(_T("Parsed lines: %u, Found IP ranges: %u, Duplicate: %u, Merged: %u"), iLine, iFoundRanges, iDuplicate, iMerged);
		}
	}

	EMULE_CATCH

	return m_IPList.GetCount();
}
/////////////////////////////////////////////////////////////////////////////////////////////
static int OptimReverseFind(const CString &strProc, TCHAR cCh)
{
	const TCHAR *pcBeg = strProc.GetString();
	const TCHAR *pcEnd = pcBeg + strProc.GetLength();

	while (--pcEnd >= pcBeg)
	{
		if (*pcEnd == cCh)
			break;
	}
	return static_cast<int>(pcEnd - pcBeg);
}
/////////////////////////////////////////////////////////////////////////////////////////////
bool CIPFilter::ParseFilterLine1(CString &rstrBuffer, uint32 &rdwIP1, uint32 &rdwIP2, UINT &ruiLevel) const
{
	TCHAR strDesc[256];
	memzero(strDesc, sizeof(strDesc));
	UINT u1, u2, u3, u4, u5, u6, u7, u8, uiLevel = DFLT_FILTER_LEVEL;
	int iItems = _stscanf(rstrBuffer, _T("%u.%u.%u.%u - %u.%u.%u.%u , %u , %255c"), &u1, &u2, &u3, &u4, &u5, &u6, &u7, &u8, &uiLevel, strDesc);
	if (iItems < 8)
		return false;

	rdwIP1 = ((u1 & 0xFF) << 24) | ((u2 & 0xFF) << 16) | ((u3 & 0xFF) << 8) | (u4 & 0xFF);
	rdwIP2 = ((u5 & 0xFF) << 24) | ((u6 & 0xFF) << 16) | ((u7 & 0xFF) << 8) | (u8 & 0xFF);

	if (iItems <= 9)
	{
		if (iItems == 8)
			uiLevel = DFLT_FILTER_LEVEL;	// set default level
		ruiLevel = uiLevel;
		rstrBuffer = _T("");
		return true;	 // no description available
	}
	ruiLevel = uiLevel;

	ASSERT( iItems == 10 );
	rstrBuffer = strDesc;
	rstrBuffer.Trim();

	return true;
}
/////////////////////////////////////////////////////////////////////////////////////////////
bool CIPFilter::ParseFilterLine2(CString &rstrBuffer, uint32 &rdwIP1, uint32 &rdwIP2, UINT &ruiLevel) const
{
	int iPos = ::OptimReverseFind(rstrBuffer, _T(':'));
	if (iPos <= 0)
		return false;

	const TCHAR *pcIPRange = rstrBuffer.GetString() + iPos + 1;
	UINT u1, u2, u3, u4, u5, u6, u7, u8;
	if (_stscanf(pcIPRange, _T("%u.%u.%u.%u - %u.%u.%u.%u"), &u1, &u2, &u3, &u4, &u5, &u6, &u7, &u8) != 8)
		return false;

	rdwIP1 = ((u1 & 0xFF) << 24) | ((u2 & 0xFF) << 16) | ((u3 & 0xFF) << 8) | (u4 & 0xFF);
	rdwIP2 = ((u5 & 0xFF) << 24) | ((u6 & 0xFF) << 16) | ((u7 & 0xFF) << 8) | (u8 & 0xFF);

	ruiLevel = DFLT_FILTER_LEVEL;

	rstrBuffer.Truncate(iPos);
	rstrBuffer.Replace(_T("PGIPDB"), _T(""));
	rstrBuffer.Trim();

	return true;
}
/////////////////////////////////////////////////////////////////////////////////////////////
void CIPFilter::RemoveAllIPs()
{
	EMULE_TRY

	for (int i = 0; i < m_IPList.GetCount(); i++)
		delete m_IPList[i];

	m_IPList.RemoveAll();
	m_pLastHit = NULL;

	EMULE_CATCH
}
/////////////////////////////////////////////////////////////////////////////////////////////
void CIPFilter::RemoveOldIPs()
{
	EMULE_TRY

	uint32	dwTime, dwCurTick = ::GetTickCount();
	hash_map<uint32, uint32>::iterator	it;

	for (it = iplist_temporary.begin(); it != iplist_temporary.end(); ++it)
	{
		dwTime = it->second;
		if ((dwCurTick - dwTime) > DFLT_TICKS_TO_BAN)
			iplist_temporary.erase(it->first);
	}

	EMULE_CATCH
}
/////////////////////////////////////////////////////////////////////////////////////////////
bool CIPFilter::IsFiltered(uint32 dwIP)
{
	return (IsFiltered(dwIP, g_App.m_pPrefs->GetIPFilterLevel()) || IsTemporaryFiltered(dwIP));
}
/////////////////////////////////////////////////////////////////////////////////////////////
static int __cdecl CmpSIPFilterByAddr(const void *pvKey, const void *pvElement)
{
	EMULE_TRY

	uint32	dwIP = *(uint32*)pvKey;
	const SIPFilter	*pIPFilter = *(SIPFilter**)pvElement;

	if (dwIP < pIPFilter->dwStart)
		return -1;

	if (dwIP > pIPFilter->dwEnd)
		return 1;

	EMULE_CATCH

	return 0;
}
/////////////////////////////////////////////////////////////////////////////////////////////
bool CIPFilter::IsFiltered(uint32 dwIP, UINT uiLevel)
{
	EMULE_TRY

	if (m_IPList.GetCount() == 0 || dwIP == 0)
		return false;

	dwIP = htonl(dwIP);

	// to speed things up we use a binary search
	//	*)	the IP filter list must be sorted by IP range start addresses
	//	*)	the IP filter list is not allowed to contain overlapping IP ranges (see also the IP range merging code when
	//		loading the list)
	//	*)	the filter 'level' is ignored during the binary search and is evaluated only for the found element
	//
	// TODO: this can still be improved even more:
	//	*)	use a pre assembled list of IP ranges which contains only the IP ranges for the currently used filter level
	//	*)	use a dumb plain array for storing the IP range structures. this will give more cach hits when processing
	//		the list. but(!) this would require to also use a dumb SIPFilter structure (don't use data items with ctors).
	//		otherwise the creation of the array would be rather slow.
	SIPFilter** ppFound = (SIPFilter**)bsearch(&dwIP, m_IPList.GetData(), m_IPList.GetCount(), sizeof(m_IPList[0]), CmpSIPFilterByAddr);
	if (ppFound && (*ppFound)->uiLevel < uiLevel)
	{
#if 0
		(*ppFound)->uiHits++;
#endif
		m_pLastHit = *ppFound;
		m_strTemporaryLastHit = _T("Not available");
		return true;
	}

	EMULE_CATCH

	return false;
}
/////////////////////////////////////////////////////////////////////////////////////////////
bool CIPFilter::IsTemporaryFiltered(uint32 dwIP)
{
	EMULE_TRY

	if (iplist_temporary.size() == 0 || dwIP == 0)
		return false;

	hash_map<uint32, uint32>::iterator it = iplist_temporary.find(dwIP);

	if (it != iplist_temporary.end())
	{
		uint32 dwTicksBanned = ::GetTickCount() - it->second;

		if (dwTicksBanned > DFLT_TICKS_TO_BAN)
		{
			iplist_temporary.erase(it->first);
			return false;
		}
		else
		{
			m_strTemporaryLastHit.Format(_T("temporarily blacklisted (%s remaining)"), CastSecondsToHM((DFLT_TICKS_TO_BAN - dwTicksBanned)/1000));
			m_pLastHit = NULL;
			return true;
		}
	}
	else
		return false;

	EMULE_CATCH

	return false;
}
/////////////////////////////////////////////////////////////////////////////////////////////
LPCTSTR CIPFilter::GetLastHit() const
{
	return m_pLastHit ? m_pLastHit->strDesc : m_strTemporaryLastHit;
}
/////////////////////////////////////////////////////////////////////////////////////////////
BOOL CIPFilter::DownloadIPFilter()
{
	EMULE_TRY

	bool bSuccess = false;
	CString strIPFilterURL = g_App.m_pPrefs->GetIPFilterURL();

	if (!strIPFilterURL.IsEmpty())
	{
		CHttpDownloadDlg dlgDownload;

		dlgDownload.m_strInitializingTitle	= GetResString(IDS_IPFILTER_HTTP_TITLE);
		dlgDownload.m_nIDPercentage			= IDS_IPFILTER_HTTP_PERCENTAGE;
		dlgDownload.m_sURLToDownload		= strIPFilterURL;
		dlgDownload.m_sFileToDownloadInto	= g_App.m_pPrefs->GetConfigDir() + DFLT_IPFILTER_FILENAME;

		CString strBackupFilePath = dlgDownload.m_sFileToDownloadInto + _T(".bak");

	//	Removes any existing backup file
		_tremove(strBackupFilePath);
	//	Backups ipfilter.dat in case the download fails
		MoveFile(dlgDownload.m_sFileToDownloadInto, strBackupFilePath);

		if (dlgDownload.DoModal() != IDOK)
		{
			g_App.AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_IPFILTER_UPDATE_ERROR));

		//	Restores original file
			_tremove(dlgDownload.m_sFileToDownloadInto);
			MoveFile(strBackupFilePath, dlgDownload.m_sFileToDownloadInto);
		}
		else
		{
		//	Delete backup file
			_tremove(strBackupFilePath);
			bSuccess = true;
			LoadFromDefaultFile();
		}
	}

	return bSuccess;

	EMULE_CATCH

	return FALSE;
}
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
